"""TTY-rendering enabled progress bar."""
from dataclasses import dataclass, field, InitVar
from typing import Optional

from ..progress import ProgressBar

from .position import Position

from .console import get_cursor_position, at_position


@dataclass
class TTYProgressBar(ProgressBar):
    """Progress bar capable of "drawing" itself to an ANSI-escape enabled TTY.

    Args:
        position: The position where the progress bar should be rendered.
                  If ommitted, uses the current cursor position.
    """
    position: InitVar[Optional[Position]] = None
    _position: Position = field(init=False)

    def __post_init__(self, starting_progress, position):
        super().__post_init__(starting_progress)
        self._position = (
            position if position is not None else get_cursor_position()
        )

    def draw(self):
        """Draw the progress bar to the tty."""
        with at_position(self._position):
            print(self.render(), end='', flush=True)



"""Main logic behind progress bar rendering."""
from dataclasses import dataclass, field, InitVar

from .errors import ProgressValueError


@dataclass
class _ProgressBarData:
    """Work around dataclasses and properties not being in good speaking terms.

    This makes `dataclass` read the properties and generate the proper method
    without thinking the default value should be the `property` instance.
    """
    size: int
    total: int
    characters: str = ' ⠁⠃⠇⡇⣇⣧⣷⣿'
    starting_progress: InitVar[int] = 0

    _steps: int = field(init=False, repr=False)
    _resolution: int = field(init=False, repr=False)
    _progress: int = field(init=False, repr=False)

    def _steps_and_resolution(self):
        steps = len(self.characters) - 1
        resolution = self.size * (steps)

        return steps, resolution

    def __post_init__(self, starting_progress):
        self._steps, self._resolution = self._steps_and_resolution()
        self._progress = starting_progress


class _ProgressBarValidation(_ProgressBarData):
    """Separate validators from main class logic."""
    @property
    def progress(self) -> int:
        """Get the current progress of the progress bar.

        Returns:
            The current progress (out of self.total).
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        """Set the current progress of the progress bar.

        Raises an error if value is less than zero or over self.total.

        Returns:
            The current progress (out of self.total).
        """
        if self.total < value < 0:
            raise ProgressValueError(
                f'Progress cannot be set higher than {self.total}. '
                f'Value was {value}.'
            )

        self._progress = value


class ProgressBar(_ProgressBarValidation):
    r"""A text-based progress bar.

    The progress bar is composed by three parts, as seen below in between '|':

    |⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣧                      |
      \    solid    / | \        empty     /
                    leading

    The solid part represents everything in the progress bar that won't change
    anymore.

    The leading character can progress through a sequence of symbols,
    artificially raising the progress bar resolution.

    The empty part is composed of characters that represent progress that
    hasn't yet been reached.

    Args:
        size: The progress bar size in characters.
        total: The total progress capacity of the bar in arbitrary units.
        progress: The starting progress of the progress bar.
        characters: The characters to use for the bar. The first one is the
                    empty one, the last one is the solid one. Everything in
                    between represents the progression of the leading
                    character, from left to right.
    """
    def _filled_amount(self):
        full_ratio = self.progress/self.total

        return int(self._resolution * full_ratio)

    def _solid_size(self, filled):
        return max(0, filled - 1) // self._steps

    def _leading_char_index(self, filled):
        if filled == 0:
            return 0

        return (filled - 1) % self._steps + 1

    def _solid_and_leading(self):
        filled = self._filled_amount()

        solid = self._solid_size(filled) * self.characters[-1]
        leading = self.characters[self._leading_char_index(filled)]

        return solid, leading

    def render(self) -> str:
        """Render the progress bar to a string.

        Returns:
            The rendered progress bar.
        """
        solid, leading = self._solid_and_leading()
        empty = self.characters[0]
        full = f'{solid}{leading}'
        bar = f'{full:{empty}<{self.size}}'

        return bar



"""Basic cursor position representation."""
from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class Position:
    """Represents a position in a text terminal.

    By usual convention for the domain, uses (line, column) notation.
    """
    y: int
    x: int

    def __add__(self, other: Position) -> Position:
        return Position(self.y + other.y, self.x + other.x)

    def __mul__(self, scalar: int) -> Position:
        return Position(self.y * scalar, self.x * scalar)

    def __sub__(self, other: Position) -> Position:
        return self + (other * -1)



"""Error classes for Proggy."""


class ProggyError(Exception):
    """Base class for all Proggy errors."""


class ProgressValueError(ProggyError):
    """Represent a value error during the manipulation of a progress bar.

    E.g.: Progress being set higher than total.
    """



"""All funcionalities related to generating and interpreting ANSI escapes."""
from enum import Enum
from typing import BinaryIO, List, Optional

from .position import Position


class ANSIEscapeType(Enum):
    """Suffixes which define ANSI escape sequence types."""
    CursorPosition = 'H'
    DeviceStatusReport = '6n'


def ansi_escape(kind: ANSIEscapeType, *, values: Optional[List[str]] = None):
    """Generate an ANSI escape string."""
    values = values if values is not None else []

    values_str = ';'.join(values)
    return f'\u001b[{values_str}{kind.value}'


device_status_report_escape = ansi_escape(ANSIEscapeType.DeviceStatusReport)


def set_cursor_position_escape(position: Position):
    """Generate the escape sequence to set the cursor position."""
    return ansi_escape(
        ANSIEscapeType.CursorPosition,
        values=[str(x) for x in [position.y, position.x]],
    )


def read_ansi_response(bytestream: BinaryIO):
    """Read an ANSI response from a bytestream (usually stdin)."""
    buf = bytearray()

    bytestream.read(2)
    while not buf.endswith(b'R'):
        buf += bytestream.read(1)

    return buf[:-1].split(b';')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['proggy', 'proggy.tty']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'proggy',
    'version': '0.2.0',
    'description': 'A progress bar generator.',
    'long_description': "Proggy\n======\n\nProgressively progressing through progress bar generation.\n\nProggy generates text-based progress bars. Mildly inspired by Rust's\n[indicatif](https://github.com/mitsuhiko/indicatif).\n\nProggy only renders progress bars to a string. Displaying them is, as of now,\nnot handled and left to the user.\n\nExamples\n--------\n\n### API usage\n\n```\n>>> from proggy import ProgressBar\n>>> pb = ProgressBar(30, 100, progress=75)\n>>> pb.render()\n'⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡇       '\n```\n\n### CLI output\n\n`test.py`:\n```python\nimport time\n\nfrom proggy import ProgressBar\n\n\np = ProgressBar(size=30, total=100)\n\np.render()\n\nfor i in range(100):\n    time.sleep(0.1)\n    p.progress += 1\n    print(f'\\u001b[{p.size}D{p.render()}', end='', flush=True)\n```\n\nOutput:\n\n![test.py output](gif/test.gif)\n",
    'author': 'Tarcísio Eduardo Moreira Crocomo',
    'author_email': 'tarcisioe@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Console and cursor control."""
from contextlib import contextmanager
from sys import stdin
from termios import tcgetattr, tcsetattr, TCSANOW
from tty import setcbreak
from typing import Optional


from .position import Position
from .ansi import (
    read_ansi_response, device_status_report_escape, set_cursor_position_escape
)


@contextmanager
def cbreak():
    """Set cbreak on TTY and reset after."""
    attr = tcgetattr(stdin)
    setcbreak(stdin, TCSANOW)
    yield
    tcsetattr(stdin, TCSANOW, attr)


def try_get_cursor_position() -> Optional[Position]:
    """Try to get the cursor position.

    Can fail if there was user input during execution.

    Returns:
        On success, the current cursor position, else None.
    """
    with cbreak():
        print(device_status_report_escape, end='', flush=True)
        y, x = read_ansi_response(stdin.buffer)

    try:
        return Position(int(y), int(x))
    except ValueError:
        return None


def get_cursor_position() -> Position:
    """Get the current cursor position in the TTY.

    Returns:
        The current cursor position.
    """
    position = None

    while position is None:
        position = try_get_cursor_position()

    return position


def set_cursor_position(position: Position):
    """Set the cursor to a given position in the TTY.

    Args:
        position: Where to set the cursor.
    """
    print(set_cursor_position_escape(position), end='', flush=True)


@contextmanager
def at_position(position: Position):
    """Set the cursor position to a given position and move it back after.

    Args:
        position: Where to set the cursor.
    """
    old = get_cursor_position()
    set_cursor_position(position)
    yield
    set_cursor_position(old)


def reserve_space(n_lines: int):
    """Reserve n lines in the console and return the position of the first one.

    Leaves the cursor in the line after the reserved lines.

    Args:
        n_lines: How many lines to reserve.
    """
    print('\n' * (n_lines), end='', flush=True)
    return get_cursor_position() - Position(n_lines, 0)
